@@ -126,7 +126,7 @@ class PackageTest(QuiltTestCase):
             pkg = Package.load(fd)
             assert test_file.resolve().as_uri() == pkg['bar'].physical_keys[0]
 
-    @patch('quilt3.Package.browse', lambda name, registry, top_hash: Package())
+    @patch('quilt3.Package._browse', lambda name, registry, top_hash: Package())
     def test_default_install_location(self):
         """Verify that pushes to the default local install location work as expected"""
         with patch('quilt3.Package._materialize') as materialize_mock:
@@ -157,52 +157,133 @@ class PackageTest(QuiltTestCase):
         assert sorted(original_set, key=lambda k: k.get('logical_key', 'manifest')) \
             == sorted(written_set, key=lambda k: k.get('logical_key', 'manifest'))
 
-    def test_browse_package_from_registry(self):
-        """ Verify loading manifest locally and from s3 """
-        with patch('quilt3.Package._from_path') as pkgmock:
-            registry = LOCAL_REGISTRY.resolve().as_uri()
-            pkg = Package()
-            pkgmock.return_value = pkg
-            top_hash = pkg.top_hash
+    def test_remote_browse(self):
+        """ Verify loading manifest from s3 """
+        registry = 's3://test-bucket'
 
-            pkg = Package.browse('Quilt/nice-name', top_hash=top_hash)
-            assert '{}/.quilt/packages/{}'.format(registry, top_hash) \
-                    in [x[0][0] for x in pkgmock.call_args_list]
+        top_hash = 'abcdefgh' * 8
 
-            pkgmock.reset_mock()
+        # Make the first request.
 
-            with patch('quilt3.packages.get_bytes') as dl_mock:
-                dl_mock.return_value = top_hash.encode('utf-8')
-                pkg = Package.browse('Quilt/nice-name')
-                assert registry + '/.quilt/named_packages/Quilt/nice-name/latest' \
-                        == dl_mock.call_args_list[0][0][0]
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(top_hash.encode()),
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Key': '.quilt/named_packages/Quilt/test/latest',
+            }
+        )
 
-            assert '{}/.quilt/packages/{}'.format(registry, top_hash) \
-                    in [x[0][0] for x in pkgmock.call_args_list]
-            pkgmock.reset_mock()
+        self.s3_stubber.add_response(
+            method='head_object',
+            service_response={
+                'VersionId': 'v1',
+                'ContentLength': REMOTE_MANIFEST.stat().st_size,
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Key': f'.quilt/packages/{top_hash}',
+            }
+        )
 
-            remote_registry = 's3://asdf/foo'
-            # remote load
-            pkg = Package.browse('Quilt/nice-name', registry=remote_registry, top_hash=top_hash)
-            assert '{}/.quilt/packages/{}'.format(remote_registry, top_hash) \
-                    in [x[0][0] for x in pkgmock.call_args_list]
-            pkgmock.reset_mock()
-            pkg = Package.browse('Quilt/nice-name', top_hash=top_hash, registry=remote_registry)
-            assert '{}/.quilt/packages/{}'.format(remote_registry, top_hash) \
-                    in [x[0][0] for x in pkgmock.call_args_list]
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(REMOTE_MANIFEST.read_bytes()),
+                'ContentLength': REMOTE_MANIFEST.stat().st_size,
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Key': f'.quilt/packages/{top_hash}',
+            }
+        )
 
-            pkgmock.reset_mock()
-            with patch('quilt3.packages.get_bytes') as dl_mock:
-                dl_mock.return_value = top_hash.encode('utf-8')
-                pkg = Package.browse('Quilt/nice-name', registry=remote_registry)
-            assert '{}/.quilt/packages/{}'.format(remote_registry, top_hash) \
-                    in [x[0][0] for x in pkgmock.call_args_list]
+        pkg = Package.browse('Quilt/test', registry=registry)
+        assert 'foo' in pkg
+
+        # Make the second request. Gets "latest" - but the rest should be cached.
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(top_hash.encode()),
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Key': '.quilt/named_packages/Quilt/test/latest',
+            }
+        )
+
+        pkg2 = Package.browse('Quilt/test', registry=registry)
+        assert 'foo' in pkg2
+
+        # Make another request with a top hash. Everything should be cached.
+
+        pkg3 = Package.browse('Quilt/test', top_hash=top_hash, registry=registry)
+        assert 'foo' in pkg3
+
+        # Make a request with a short hash.
+
+        self.s3_stubber.add_response(
+            method='list_objects_v2',
+            service_response={
+                'Contents': [
+                    {
+                        'Key': f'.quilt/packages/{top_hash}',
+                        'Size': 64,
+                    },
+                    {
+                        'Key': f'.quilt/packages/{"a" * 64}',
+                        'Size': 64,
+                    }
+                ]
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Prefix': '.quilt/packages/',
+            }
+        )
+
+        pkg3 = Package.browse('Quilt/test', top_hash='abcdef', registry=registry)
+        assert 'foo' in pkg3
+
+        # Make a request with a bad short hash.
+
+        with self.assertRaises(QuiltException):
+            Package.browse('Quilt/test', top_hash='abcde', registry=registry)
+        with self.assertRaises(QuiltException):
+            Package.browse('Quilt/test', top_hash='a' * 65, registry=registry)
+
+        # Make a request with a non-existant short hash.
+
+        self.s3_stubber.add_response(
+            method='list_objects_v2',
+            service_response={
+                'Contents': [
+                    {
+                        'Key': f'.quilt/packages/{top_hash}',
+                        'Size': 64,
+                    },
+                    {
+                        'Key': f'.quilt/packages/{"a" * 64}',
+                        'Size': 64,
+                    }
+                ]
+            },
+            expected_params={
+                'Bucket': 'test-bucket',
+                'Prefix': '.quilt/packages/',
+            }
+        )
+
+        with self.assertRaises(QuiltException):
+            Package.browse('Quilt/test', top_hash='123456', registry=registry)
 
-            # registry failure case
-            with patch('quilt3.packages.get_from_config',
-                       return_value=fix_url(os.path.dirname(__file__))):
-                with pytest.raises(FileNotFoundError):
-                    Package.browse('Quilt/nice-name')
 
     def test_remote_install(self):
         """Verify that installing from a local package works as expected."""
@@ -216,7 +297,7 @@ class PackageTest(QuiltTestCase):
             pkg.build('Quilt/nice-name')
 
             with patch('quilt3.Package._materialize') as materialize_mock, \
-                patch('quilt3.Package.build') as build_mock:
+                patch('quilt3.Package._build') as build_mock:
                 materialize_mock.return_value = pkg
                 dest_registry = quilt3.util.get_from_config('default_local_registry')
 
@@ -664,7 +745,7 @@ class PackageTest(QuiltTestCase):
         # Test that push cleans up the temporary files, if and only if the serialization_location was not set
         with patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call), \
             patch('quilt3.Package._materialize') as materialize_mock, \
-            patch('quilt3.Package.build') as build_mock:
+            patch('quilt3.Package._build') as build_mock:
             materialize_mock.return_value = pkg
 
             pkg.push('Quilt/test_pkg_name', 's3://test-bucket')
@@ -939,7 +1020,7 @@ class PackageTest(QuiltTestCase):
         """ Verify commit messages populate correctly on push."""
         with patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call), \
             patch('quilt3.Package._materialize') as materialize_mock, \
-            patch('quilt3.Package.build') as build_mock:
+            patch('quilt3.Package._build') as build_mock:
             with open(REMOTE_MANIFEST) as fd:
                 pkg = Package.load(fd)
             materialize_mock.return_value = pkg
@@ -1060,8 +1141,8 @@ class PackageTest(QuiltTestCase):
 
 
     def test_import(self):
-        with patch('quilt3.Package.browse') as browse_mock, \
-            patch('quilt3.imports.list_packages') as list_packages_mock:
+        with patch('quilt3.Package._browse') as browse_mock, \
+            patch('quilt3.imports._list_packages') as list_packages_mock:
             browse_mock.return_value = quilt3.Package()
             list_packages_mock.return_value = ['foo/bar', 'foo/baz']
 
@@ -1109,6 +1190,139 @@ class PackageTest(QuiltTestCase):
         with pytest.raises(QuiltException):
             pkg.set('s3://foo/..', LOCAL_MANIFEST)
 
+    def test_install(self):
+        # Manifest
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(b'abcdef'),
+            },
+            expected_params={
+                'Bucket': 'my-test-bucket',
+                'Key': '.quilt/named_packages/Quilt/Foo/latest',
+            }
+        )
+
+        self.s3_stubber.add_response(
+            method='head_object',
+            service_response={
+                'VersionId': 'v1',
+                'ContentLength': REMOTE_MANIFEST.stat().st_size,
+            },
+            expected_params={
+                'Bucket': 'my-test-bucket',
+                'Key': '.quilt/packages/abcdef',
+            }
+        )
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(REMOTE_MANIFEST.read_bytes()),
+                'ContentLength': REMOTE_MANIFEST.stat().st_size,
+            },
+            expected_params={
+                'Bucket': 'my-test-bucket',
+                'Key': '.quilt/packages/abcdef',
+            }
+        )
+
+        # Objects
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(b'a,b,c'),
+            },
+            expected_params={
+                'Bucket': 'my_bucket',
+                'Key': 'my_data_pkg/bar.csv',
+            }
+        )
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(b'Hello World!'),
+            },
+            expected_params={
+                'Bucket': 'my_bucket',
+                'Key': 'my_data_pkg/baz/bat',
+            }
+        )
+
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO('💩'.encode()),
+            },
+            expected_params={
+                'Bucket': 'my_bucket',
+                'Key': 'my_data_pkg/foo',
+            }
+        )
+
+        with patch('quilt3.data_transfer.s3_transfer_config.max_request_concurrency', 1):
+            Package.install('Quilt/Foo', registry='s3://my-test-bucket', dest='package')
+
+        p = Package.browse('Quilt/Foo')
+
+        assert p['foo'].get() == 's3://my_bucket/my_data_pkg/foo'
+
+        # Check that the cache works.
+        local_path = pathlib.Path(p['foo'].get_cached_path())
+        assert local_path == pathlib.Path.cwd() / 'package/foo'
+        assert local_path.read_text('utf8') == '💩'
+
+        # Test that get_bytes and get_as_text works
+        assert p['foo'].get_bytes().decode("utf-8") == '💩'
+        assert p['foo'].get_as_string() == '💩'
+
+        # Check that moving the file invalidates the cache...
+        local_path.rename('foo2')
+        assert p['foo'].get_cached_path() is None
+
+        # ...but moving it back fixes it.
+        pathlib.Path('foo2').rename(local_path)
+        assert p['foo'].get_cached_path() == str(local_path)
+
+        # Check that changing the contents invalidates the cache.
+        local_path.write_text('omg')
+        assert p['foo'].get_cached_path() is None
+
+        # Check that installing the package again reuses the cached manifest and two objects - but not "foo".
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO(b'abcdef'),
+            },
+            expected_params={
+                'Bucket': 'my-test-bucket',
+                'Key': '.quilt/named_packages/Quilt/Foo/latest',
+            }
+        )
+        self.s3_stubber.add_response(
+            method='get_object',
+            service_response={
+                'VersionId': 'v1',
+                'Body': BytesIO('💩'.encode()),
+            },
+            expected_params={
+                'Bucket': 'my_bucket',
+                'Key': 'my_data_pkg/foo',
+            }
+        )
+
+        with patch('quilt3.data_transfer.s3_transfer_config.max_request_concurrency', 1):
+            Package.install('Quilt/Foo', registry='s3://my-test-bucket', dest='package/')
+
     def test_rollback(self):
         p = Package()
         p.set('foo', DATA_DIR / 'foo.txt')
@@ -1128,7 +1342,29 @@ class PackageTest(QuiltTestCase):
         assert 'foo' in Package.browse('quilt/tmp')
 
         with self.assertRaises(QuiltException):
-            Package.rollback('quilt/tmp', LOCAL_REGISTRY, '12345')
+            Package.rollback('quilt/tmp', LOCAL_REGISTRY, '12345678' * 8)
 
         with self.assertRaises(QuiltException):
             Package.rollback('quilt/blah', LOCAL_REGISTRY, good_hash)
+
+
+    def test_verify(self):
+        pkg = Package()
+
+        pkg.set('foo', b'Hello, World!')
+        pkg.build('quilt/test')
+
+        Package.install('quilt/test', LOCAL_REGISTRY, dest='test')
+        assert pkg.verify('test')
+
+        Path('test/blah').write_text('123')
+        assert not pkg.verify('test')
+        assert pkg.verify('test', extra_files_ok=True)
+
+        Path('test/foo').write_text('123')
+        assert not pkg.verify('test')
+        assert not pkg.verify('test', extra_files_ok=True)
+
+        Path('test/foo').write_text('Hello, World!')
+        Path('test/blah').unlink()
+        assert pkg.verify('test')
